# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os

from django.core.files.storage import FileSystemStorage
from django.core.validators import MinLengthValidator
from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.deconstruct import deconstructible

import django_drf_filepond.drf_filepond_settings as local_settings


FILEPOND_UPLOAD_TMP = getattr(local_settings, 'UPLOAD_TMP',
                              os.path.join(
                                  local_settings.BASE_DIR,'filepond_uploads'))


@deconstructible
class FilePondUploadSystemStorage(FileSystemStorage):
    """
    Subclass FileSystemStorage to prevent creation of new migrations when
    using a file store location passed to FileSystemStorage using the
    location attribute. Instead the location is applied dynamically on
    creation of the subclass avoiding detection of changes by the migration
    system.

    Addresses #13. Fix is based on fix for similar issue in
    https://github.com/julen/pootle/commit/fd7800050172549e9f31544843b986691290ddc2
    """

    def __init__(self, **kwargs):
        kwargs.update({
            'location': FILEPOND_UPLOAD_TMP,
        })
        super(FilePondUploadSystemStorage, self).__init__(**kwargs)


storage = FilePondUploadSystemStorage()


LOG = logging.getLogger(__name__)


def get_upload_path(instance, filename):
    return os.path.join(instance.upload_id, filename)

class TemporaryUpload(models.Model):
    
    FILE_DATA = 'F'
    URL = 'U'
    UPLOAD_TYPE_CHOICES = (
        (FILE_DATA, 'Uploaded file data'),
        (URL, 'Remote file URL'),
    )
    
    # The unique ID returned to the client and the name of the temporary 
    # directory created to hold file data
    upload_id = models.CharField(primary_key=True, max_length=22, 
                               validators=[MinLengthValidator(22)])
    # The unique ID used to store the file itself
    file_id = models.CharField(max_length=22, 
                               validators=[MinLengthValidator(22)])
    file = models.FileField(storage=storage, upload_to=get_upload_path)
    upload_name = models.CharField(max_length=512)
    uploaded = models.DateTimeField(auto_now_add=True)
    upload_type = models.CharField(max_length = 1, 
                                   choices=UPLOAD_TYPE_CHOICES)
    
    def get_file_path(self):
        return self.file.path

class StoredUpload(models.Model):
        
    # The unique upload ID assigned to this file when it was originally 
    # uploaded (or retrieved from a remote URL) 
    upload_id = models.CharField(primary_key=True, max_length=22, 
                               validators=[MinLengthValidator(22)])
    # The file name and path (relative to the base file store directory 
    # as set by DJANGO_DRF_FILEPOND_FILE_STORE_PATH).
    file_path = models.CharField(max_length=2048)
    uploaded = models.DateTimeField()
    stored = models.DateTimeField(auto_now_add=True)
    
    def get_absolute_file_path(self):
        return os.path.join(local_settings.FILE_STORE_PATH, self.file_path)
    
# When a TemporaryUpload record is deleted, we need to delete the 
# corresponding file from the filesystem by catching the post_delete signal.
@receiver(post_delete, sender=TemporaryUpload)
def delete_temp_upload_file(sender, instance, **kwargs):
    # Check that the file parameter for the instance is not None
    # and that the file exists and is not a directory! Then we can delete it
    LOG.debug('*** post_delete signal handler called. Deleting file.')
    if instance.file:
        if (os.path.exists(instance.file.path) and 
            os.path.isfile(instance.file.path)):
            os.remove(instance.file.path)
    
    if local_settings.DELETE_UPLOAD_TMP_DIRS:
        file_dir = os.path.join(storage.location, instance.upload_id)
        if(os.path.exists(file_dir) and os.path.isdir(file_dir)):
            os.rmdir(file_dir)
            LOG.debug('*** post_delete signal handler called. Deleting temp '
                      'dir that contained file.')




# This module contains API functions exported through the top-level
# django_drf_filepond module.
#
# store_upload: used to move an upload from temporary storage to a permanent
#               storage location. This requires that you have the 
#               DJANGO_DRF_FILEPOND_FILE_STORE_PATH setting set in your 
#               application's settings.py file.
#
import logging
import os
import shutil

import django_drf_filepond.drf_filepond_settings as local_settings
from django.core.exceptions import ImproperlyConfigured
import re
import shortuuid
from django_drf_filepond.models import TemporaryUpload, StoredUpload

LOG = logging.getLogger(__name__)

# Store the temporary upload represented by upload_id to the specified  
# destination_file_path under the defined file store location as specified by 
# the DJANGO_DRF_FILEPOND_FILE_STORE_PATH configuration setting. 
# Files stored using this approach can subsequently be retrieved using the  
# load method defined in by the filepond server spec by using either the 
# 22-char upload_id or the value provided to the destination_file_path 
# parameter as a query string parameter using the "id" key.
def store_upload(upload_id, destination_file_path):
    if ((not hasattr(local_settings, 'FILE_STORE_PATH')) 
        or
        (not local_settings.FILE_STORE_PATH)
        or 
        (not os.path.exists(local_settings.FILE_STORE_PATH))
        or
        (not os.path.isdir(local_settings.FILE_STORE_PATH))):
        
        raise ImproperlyConfigured('A required setting is missing in your '
                                   'application configuration.')
    
    file_path_base = local_settings.FILE_STORE_PATH
    
    if not file_path_base or file_path_base == '':
        raise ValueError('The FILE_STORE_PATH is not set to a directory.')
    
    id_fmt = re.compile('^([%s]){22}$' % (shortuuid.get_alphabet()))
        
    if not id_fmt.match(upload_id):
        LOG.error('The provided upload ID <%s> is of an invalid format.'
                  % upload_id)
        raise ValueError('The provided upload ID is of an invalid format.')
            
    if not destination_file_path or destination_file_path == '':
        raise ValueError('No destination file path provided.')
    
    if destination_file_path.startswith(os.sep):
        destination_file_path = destination_file_path[1:]
        
    try:
        tu = TemporaryUpload.objects.get(upload_id=upload_id)
    except TemporaryUpload.DoesNotExist:
        raise ValueError('Record for the specified upload_id doesn\'t exist')
    
    target_dir = os.path.join(file_path_base, 
                              os.path.dirname(destination_file_path))
    if destination_file_path.endswith(os.sep):
        # Assume a directory was provided, get the file name from tu and 
        # add this to the provided path.
        destination_file_path += tu.upload_name
        
    target_file_path = os.path.join(file_path_base, destination_file_path)
    
    # Check we're not about to overwrite anything
    if os.path.exists(target_file_path):
        LOG.error('File with specified name and path <%s> already exists' 
                  % destination_file_path)
        raise FileExistsError('The specified temporary file cannot be stored'
                              ' to the specified location - file exists.')
    
    su = StoredUpload(upload_id=tu.upload_id,  
                      file_path=destination_file_path, uploaded=tu.uploaded)
    
    try:
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        shutil.copy2(tu.get_file_path(), target_file_path)
        su.save()
        tu.delete()
    except IOError as e:
        LOG.error('Error moving temporary file to permanent storage location')
        raise e
    
    return su


'''
Setup some local application settings with local defaults.
This approach is based on the one shown in this blog post:

http://blog.muhuk.com/2010/01/26/developing-reusable-django-apps-app-settings.html#.W_Rkh-mYQ6U

******** Settings in this file are not currently being used for core library
******** use due to an issue with dynamically setting up the storage  
******** location in the FileSystemStorage class. For now, you must set 
******** DJANGO_DRF_FILEPOND_UPLOAD_TMP in your top level app settings.
******** THIS IS CURRENTLY USED ONLY FOR INTEGRATION TESTS
'''
import os
from django.conf import settings
import django_drf_filepond

_app_prefix = 'DJANGO_DRF_FILEPOND_'

# BASE_DIR is assumed to be present in the core project settings. However,
# in case it isn't, we check here and set a local BASE_DIR to the 
# installed app base directory to use as an alternative.
BASE_DIR = os.path.dirname(django_drf_filepond.__file__)
if hasattr(settings, 'BASE_DIR'):
    BASE_DIR = settings.BASE_DIR

# The location where uploaded files are temporarily stored. At present, 
# this must be a subdirectory of settings.BASE_DIR
UPLOAD_TMP = getattr(settings, _app_prefix+'UPLOAD_TMP',
                     os.path.join(BASE_DIR,'filepond_uploads'))

# Setting to control whether the temporary directories created for file 
# uploads are removed when an uploaded file is deleted
DELETE_UPLOAD_TMP_DIRS = getattr(settings, 
                                 _app_prefix+'DELETE_UPLOAD_TMP_DIRS', True)

# The file storage location used by the top-level application. This needs to 
# be set if the load endpoint is going to be used to access files that have 
# been permanently stored after being uploaded as TemporaryUpload objects.
FILE_STORE_PATH = getattr(settings, _app_prefix+'FILE_STORE_PATH', None)


from django.apps import AppConfig

import os
import logging
import django_drf_filepond.drf_filepond_settings as local_settings

LOG = logging.getLogger(__name__)

class DjangoDrfFilepondConfig(AppConfig):
    name = 'django_drf_filepond'
    verbose_name = 'FilePond Server-side API'
    
    def ready(self):
        upload_tmp = getattr(local_settings, 'UPLOAD_TMP',
                             os.path.join(
                                 local_settings.BASE_DIR,'filepond_uploads'))
        LOG.debug('Upload temp directory from top level settings: <%s>'
                  % (upload_tmp))
        
        # Check if the temporary file directory is available and if not
        # create it.
        if not os.path.exists(local_settings.UPLOAD_TMP):
            LOG.warning('Filepond app init: Creating temporary file '
                     'upload directory <%s>' % local_settings.UPLOAD_TMP)
            os.makedirs(local_settings.UPLOAD_TMP, mode=0o700)
        else:
            LOG.debug('Filepond app init: Temporary file upload '
                      'directory already exists')

        file_store = getattr(local_settings, 'FILE_STORE_PATH', None)
        if file_store:
            if not os.path.exists(file_store):
                LOG.warning('Filepond app init: Creating file store '
                            'directory <%s>...' % file_store)
                os.makedirs(file_store, mode=0o700)
            else:
                LOG.debug('Filepond app init: File store path already '
                          'exists')


"""FilePond server-side URL configuration

Based on the server-side configuration details
provided at: 
https://pqina.nl/filepond/docs/patterns/api/server/#configuration
"""
from django.conf.urls import url
from django_drf_filepond.views import ProcessView, RevertView, LoadView,\
     RestoreView, FetchView

urlpatterns = [
    url(r'^process/$', ProcessView.as_view(), name='process'),
    url(r'^revert/$', RevertView.as_view(), name='revert'),
    url(r'^load/$', LoadView.as_view(), name='load'),
    url(r'^restore/$', RestoreView.as_view(), name='restore'),
    url(r'^fetch/$', FetchView.as_view(), name='fetch')
]
